#include <Endian.h>
#include <ImageCompression.h>
#include <Movies.h>
#include <Script.h>
#include <Strings.h>

#include "EI_Image.h"
#include "EI_MakeImageDescription.h"

#include "EI_GraphicsImportVersion.h"

typedef struct {
	ComponentInstance		self;
	ComponentInstance		parent;
	ComponentInstance		delegateComponent;
} EI_GraphicsImportGlobalsRecord, *EI_GraphicsImportGlobals;

#define GRAPHICSIMPORT_BASENAME() 		EI_GraphicsImport
#define GRAPHICSIMPORT_GLOBALS() 		EI_GraphicsImportGlobals storage

#define CALLCOMPONENT_BASENAME()		GRAPHICSIMPORT_BASENAME()
#define	CALLCOMPONENT_GLOBALS()			GRAPHICSIMPORT_GLOBALS()

#define COMPONENT_UPP_PREFIX()			uppGraphicsImport
#define COMPONENT_DISPATCH_FILE			"EI_GraphicsImportDispatch.h"
#define COMPONENT_SELECT_PREFIX()  		kGraphicsImport

#define	GET_DELEGATE_COMPONENT()		(storage->delegateComponent)

#include <Components.k.h>
#include <ImageCompression.k.h>
#include <ComponentDispatchHelper.c>

pascal ComponentResult EI_GraphicsImportOpen (EI_GraphicsImportGlobals store, ComponentInstance self)
{
	ComponentResult err = noErr;

	store = (EI_GraphicsImportGlobals)NewPtrClear(sizeof(EI_GraphicsImportGlobalsRecord));
	if (err = MemError()) goto bail;

	SetComponentInstanceStorage(self, (Handle)store);

	store->self = self;
	store->parent = self;

	err = OpenADefaultComponent(GraphicsImporterComponentType, GraphicsImporterComponentType, &store->delegateComponent);
	if (err) goto bail;

	err = ComponentSetTarget(store->delegateComponent, self);
	if (err) goto bail;

bail:
	return err;
}

pascal ComponentResult EI_GraphicsImportClose (EI_GraphicsImportGlobals store, ComponentInstance self)
{
#pragma unused(self)
	if (store) {
		if (store->delegateComponent)
			CloseComponent(store->delegateComponent);

		DisposePtr((Ptr)store);
	}
	return noErr;
}


pascal ComponentResult EI_GraphicsImportVersion (EI_GraphicsImportGlobals store)
{
#pragma unused(store)
#ifndef POWERPC_QT
	return kEI_GraphicsImportVersion;
#else
	return kEI_GraphicsImportVersionPPC;
#endif
}

pascal ComponentResult EI_GraphicsImportTarget (EI_GraphicsImportGlobals store, ComponentInstance target)
{
	store->parent = target;
	return noErr;
}

pascal ComponentResult EI_GraphicsImportGetImageDescription (EI_GraphicsImportGlobals store, ImageDescriptionHandle *descOut)
{
	ComponentResult err = noErr;
	ImageFrame frame;
	unsigned long offset, size;
	short colorCount;
	Ptr colors = nil;

	if (!descOut)
		return paramErr;

	err = GraphicsImportGetDataOffsetAndSize(store->parent, &offset, &size);
	if (err) goto bail;

	err = GraphicsImportReadData(store->parent, &frame, offset, sizeof(frame));
	if (err) goto bail;

	colorCount = EndianU16_BtoN(frame.framePalettes);

	colors = NewPtr(3 * colorCount);
	if (err = MemError()) goto bail;

	err = GraphicsImportReadData(store->parent, colors, offset + sizeof(frame), 3 * colorCount);
	if (err) goto bail;

	err = EI_MakeImageDescription(&frame, colorCount, (UInt8 *)colors, descOut);
	if (err) goto bail;

bail:
	if (colors)
		DisposePtr(colors);

	return err;	
}

pascal ComponentResult EI_GraphicsImportGetDataOffsetAndSize (EI_GraphicsImportGlobals store, unsigned long *offsetOut, unsigned long *sizeOut)
{
	ComponentResult err = noErr;
	ImageHeader header;
	ImageFrame frame;
	unsigned long offset, size;

	err = GraphicsImportGetDataOffsetAndSize(store->delegateComponent, &offset, &size);
	if (err) goto bail;

	err = GraphicsImportReadData(store->parent, &header, offset, sizeof(header));
	if (err) goto bail;

	if ((EndianU16_BtoN(header.imageVersion) != 5) || (EndianU32_BtoN(header.imageFrames) < 1)) {
		err = codecBadDataErr;
		goto bail;
	}

	err = GraphicsImportReadData(store->parent, &frame, offset + sizeof(header), sizeof(frame));
	if (err) goto bail;

	if (offsetOut) *offsetOut = offset + sizeof(ImageHeader);
	if (sizeOut) *sizeOut = sizeof(ImageFrame) + (EndianU16_BtoN(frame.framePalettes) * 3) + EndianU32_BtoN(frame.frameSize);

bail:
	return err;
}

pascal ComponentResult EI_GraphicsImportValidate (EI_GraphicsImportGlobals store, Boolean *valid)
{
	ComponentResult err = noErr;
	unsigned char header[2];

	*valid = false;

	err = GraphicsImportReadData(store->parent, &header[0], 0, 2);
	if (err) goto bail;
	
	// If your image format has a "magic number" at the start, this is where you should look for it.
	
	// Electric Image files don't have a magic number, but they do have a version code, which should
	// be a big-endian two-byte integer "5".  So we'll check that. 
	
	if( ( 0 == header[0] )
	 && ( 5 == header[1] ) )
		*valid = true;

bail:
	return err;
}

pascal ComponentResult EI_GraphicsImportGetMetaData (EI_GraphicsImportGlobals store, void *userData)
{
	ComponentResult err = noErr;
	Handle	dataHandle = NewHandle(0);

	// For the sake of example we'll return a fixed string as metadata.
	char theCString[255] = "File created with Electric Image";
	StringPtr thePString = c2pstr( theCString );

	if (nil == userData)
	{
		err = paramErr;
		goto bail;
	}

	if (nil == dataHandle)
	{
		err = memFullErr;
		goto bail;
	}
	
	err = PtrAndHand(&thePString[1], dataHandle, thePString[0]);
	if (err) goto bail;
	
	err = AddUserDataText(userData, dataHandle, kUserDataTextInformation, 1, langEnglish);
	if (err) goto bail;

bail:
	if (dataHandle)
		DisposeHandle(dataHandle);

	return err;
}

pascal ComponentResult EI_GraphicsImportGetMIMETypeList(EI_GraphicsImportGlobals store, void *qtAtomContainerPtr)
{
	// Note that GetComponentResource is only available in QuickTime 3.0 or later.
	// However, it's safe to call it here because GetMIMETypeList is only defined in QuickTime 3.0 or later.
	return GetComponentResource((Component)store->self, FOUR_CHAR_CODE('mime'), 128, (Handle *)qtAtomContainerPtr);
}

#ifndef STAND_ALONE

extern void EI_GraphicsImporterRegister(void);
void EI_GraphicsImporterRegister(void)
{
	ComponentDescription foo;
	ComponentRoutineUPP componentEntryPoint = NewComponentRoutineProc(EI_GraphicsImportComponentDispatch);

  	foo.componentType = GraphicsImporterComponentType;
  	foo.componentSubType = FOUR_CHAR_CODE('EIDI');
  	foo.componentManufacturer = kAppleManufacturer;
  	foo.componentFlags = 0;
  	foo.componentFlagsMask = 0;

	RegisterComponent(&foo, (void *)componentEntryPoint, 1, 0, 0, 0);
}

#endif
