/*
 * QuickTime for Java SDK Sample Code

   Usage subject to restrictions in SDK License Agreement
 * Copyright:  1996-1999 Apple Computer, Inc.

 */
package mixer.display;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import quicktime.*;
import quicktime.app.audio.*;

import mixer.*;
import mixer.mc.*;


/** This is the base class for any set of controllers to display to the user.  It
 *  is possible to modify which buttons are presented to control a channel by subclassing
 *  this class and changing the boolean value associate with each control and/or the
 *  the listener that deals with each component.
 */
public abstract class ChannelDisplay extends JPanel {
 	protected MixerComponents mixer;
	protected JCheckBox muteButton, soloButton;
	private boolean mute, solo, balance, volume;
	private static final int CONTROLLER_WIDTH = 70;
	private static final int EDIT_HEIGHT      = 35;
	private static final int MUTE_HEIGHT      = 28;
	private static final int SOLO_HEIGHT      = 28;
	private static final int BALANCE_HEIGHT   = 18;
	private static final int VOLUME_HEIGHT    = 170;
	
	/** This constructor goes through and creates the actual layout of controls for
	 *  the mixer.  It needs to know which set of components are being mixed, and
	 *  which controls you would like to have present.  It creates the desired controls
	 *  and calls the add*Action methods to add listeners to the component.  All of these
	 *  methods have dummy implementations if, for some reason, you want the control, but
	 *  no listener.  Note that the edit button will be displayed if the media argument
	 *  says that the MixerComponents object is editable.
	 *  @param media the MixerComponents object that is being controlled
	 *  @param mute add a mute button?
	 *  @param solo add a solo button?
	 *  @param balance add a balance slider?
	 *  @param volume add a volume slider?
	 */
	protected ChannelDisplay (MixerComponents media, boolean mute, boolean solo,
					 boolean balance, boolean volume) {
		super(new GridBagLayout());
		
		mixer = media;
		this.mute = mute;
		this.solo = solo;
		this.balance = balance;
		this.volume = volume;
		
		// setup the constraints
		GridBagConstraints cons = new GridBagConstraints();
		cons.gridx = 0;
		cons.gridy = GridBagConstraints.RELATIVE;
		cons.weightx = 0.0F;
		cons.weighty = 0.0F;
		cons.insets = new Insets(1,0,0,0);
		cons.fill = GridBagConstraints.NONE;
		cons.anchor = GridBagConstraints.CENTER;
		
		// add edit button
		JComponent editBox = new JPanel();
		if (mixer.isEditable())
			editBox.add(buildEditButton());
		addJComponent(editBox, EDIT_HEIGHT, cons);
		
		// add mute button
		cons.insets = new Insets(0,0,0,0);
		JComponent muteBox = new JPanel();
		if (mute) {
			muteButton = new JCheckBox("mute");
			addMuteAction (muteButton);
			muteBox.add(muteButton);
		}
		addJComponent(muteBox, MUTE_HEIGHT, cons);

		// add solo button
		JComponent soloBox = new JPanel();
		if (solo) {
			soloButton = new JCheckBox("solo");
			addSoloAction (soloButton);
			soloBox.add(soloButton);
		}
		addJComponent(soloBox, SOLO_HEIGHT, cons);

		// add balance slider
		JComponent balanceBox = new JPanel();
		if (balance) {
			JSlider balanceSlider = new JSlider(JSlider.HORIZONTAL, -50, 50, 0);
			Dimension size = new Dimension(CONTROLLER_WIDTH, BALANCE_HEIGHT);
			balanceSlider.setPreferredSize(size);
			addBalanceAction(balanceSlider);
			balanceBox.add(balanceSlider);
		}
		addJComponent(balanceBox, BALANCE_HEIGHT + 15, cons);
				
		// add volume slider
		cons.insets = new Insets(0,0,1,0);
		JComponent volumeBox = new JPanel();
		if (volume) {
			JSlider volumeSlider = new JSlider(JSlider.VERTICAL, 0, 100, 66);
			volumeSlider.setPaintTicks(true);
			volumeSlider.setMajorTickSpacing(66);
			volumeSlider.setMinorTickSpacing(33);
			addVolumeAction(volumeSlider);
			volumeBox.add(volumeSlider);
		}
		addJComponent(volumeBox, VOLUME_HEIGHT, cons);
		
		setBorder(new LineBorder(Color.black));
	}
	
	/* This method sets the size of the component to be added to the control panel,
	 * and then add it.
	 */
	private void addJComponent (JComponent c, int height, GridBagConstraints cons) {
		Dimension d = new Dimension(CONTROLLER_WIDTH, height);
		c.setPreferredSize(d);
		c.setMaximumSize(d);
		c.setMinimumSize(d);
		add(c, cons);
	}
	
	/* The edit button is a special case over which the user has no option to leave out.
	 * Our code figures out whether or not the button belongs in the controls and adds
	 * an appropriate action listener.  Note that the makeEditComponent method of the
	 * MixerComponent is called when the Edit button is clicked.
	 */
	private JButton buildEditButton () {
		JButton editButton = new JButton("Edit");
		
		editButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				try {
					new MixerDialog (mixer.makeEditComponent());
				} catch (QTException ex) {
					ex.printStackTrace();
				}
			}
		});
		return editButton;
	}
	
	/* Dummy implementations of each of the add methods. */
	void addMuteAction(JCheckBox b) {}

	void addSoloAction(JCheckBox b) {}
	
	void addBalanceAction(JSlider b) {}
	
	void addVolumeAction(JSlider b) {}
}
